# -*- coding: utf-8 -*-

from ipaddress import ip_address
from itertools import product
from os import mkdir
from os.path import join

import numpy as np
from hermespy.modem.waveform_generator_ofdm import OFDMIdealChannelEstimation, OFDMLeastSquaresChannelEstimation, OFDMZeroForcingChannelEqualization
from hermespy.modem.waveform_single_carrier import SingleCarrierIdealChannelEstimation, SingleCarrierLeastSquaresChannelEstimation, SingleCarrierZeroForcingChannelEqualization

from hermespy.simulation import Simulation
from hermespy.modem import BitErrorEvaluator, DuplexModem, RootRaisedCosineWaveform, ChirpFSKWaveform, OFDMWaveform, FrameResource, FrameSymbolSection, FrameElement
from hermespy.tools import db2lin
from hermespy.channel import MultipathFadingChannel
from hermespy.core import ConsoleMode

__author__ = "Jan Adler"
__copyright__ = "Copyright 2022, Barkhausen Institut gGmbH"
__credits__ = ["Jan Adler"]
__license__ = "AGPLv3"
__version__ = "0.3.0"
__maintainer__ = "Jan Adler"
__email__ = "jan.adler@barkhauseninstitut.org"
__status__ = "Prototype"


sc = RootRaisedCosineWaveform(oversampling_factor=1, symbol_rate=100e6, num_data_symbols=100, num_preamble_symbols=0)
sc.channel_estimation = SingleCarrierIdealChannelEstimation()
sc.channel_equalization = SingleCarrierZeroForcingChannelEqualization()

ofdm = OFDMWaveform(oversampling_factor=1, resources=[FrameResource(1200, 0., [FrameElement('DATA')])], structure=[FrameSymbolSection(1, [0])])
ofdm.channel_estimation = OFDMIdealChannelEstimation()
ofdm.channel_equalization = OFDMZeroForcingChannelEqualization()

fsk = ChirpFSKWaveform(oversampling_factor=1, num_data_chirps=100, num_pilot_chirps=0)

waveforms = [
    ('sc', sc),
    ('ofdm', ofdm),
    ('fsk', fsk),
]

waveforms[0][0]

channel = MultipathFadingChannel([0.], [1.], [0.])
modulation_orders = [2, 4, 16, 64]

for modulation_order, (waveform_name, waveform) in product(modulation_orders, waveforms):

    simulation = Simulation()
    device = simulation.scenario.new_device()

    modem = DuplexModem()
    modem.device = device
    modem.waveform_generator = waveform
    modem.waveform_generator.modulation_order = modulation_order
    simulation.scenario.set_channel(device, device, channel)

    simulation.new_dimension('snr', [db2lin(x) for x in np.arange(-10, 20, .5)])
    simulation.add_evaluator(BitErrorEvaluator(modem, modem))
    simulation.num_samples, simulation.min_num_samples = 100000, 100000
    simulation.plot_results = True
    
    simulation.results_dir = simulation.default_results_dir()
    _ = simulation.run()
